cpdef Unitary_Grid(double complex[:,:,::1] H, double[::1] c_mins, double[::1] dcs, long[::1] c_bins)

cpdef Create_Interpolation_Cache(double complex[:,:,::1] U_grid, long[::1] grid_cum_prod, long[::1] c_bins)
cpdef Create_Sym_Interpolation_Cache(double complex[:,:,::1] U_grid2, long[::1] grid_cum_prod, long[::1] c_bins)
